/* dirmngr.c - LDAP access
 *	Copyright (C) 2002 Klarlvdalens Datakonsult AB
 *
 * This file is part of DirMngr.
 *
 * DirMngr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DirMngr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

#include <gcrypt.h>
#include <ksba.h>

#define JNLIB_NEED_LOG_LOGV
#include "dirmngr.h"
#include "crlcache.h"

#include "../assuan/assuan.h" /* needed for the malloc hooks */

/* no i18n needed */
#define N_(a) a
#define _(a) a


enum cmd_and_opt_values 
{ aNull = 0,
  oCsh		  = 'c',
  oQuiet	  = 'q',
  oSh		  = 's',
  oVerbose	  = 'v',
  
  oNoVerbose = 500,
  oOptions,
  oDebug,
  oDebugAll,
  oDebugWait,
  oNoGreeting,
  oNoOptions,
  oHomedir,
  oNoDetach,
  oNoGrab,
  oClient,
  oShutdown,
  oFlush,
  oLogFile,
  oServer,
  oBatch,
  oLDAPFile,
  oLDAPTimeout,
  oLDAPAddServers,
  oMaxReplies,
  oListCRLs,
  oLoadCRL,
aTest };



static ARGPARSE_OPTS opts[] = {
  
  { 301, NULL, 0, N_("@Options:\n ") },

  { oServer,   "server",     0, N_("run in server mode") },
  { oVerbose, "verbose",   0, N_("verbose") },
  { oQuiet,	"quiet",     0, N_("be somewhat more quiet") },
  { oSh,	"sh",        0, N_("sh-style command output") },
  { oCsh,	"csh",       0, N_("csh-style command output") },
  { oOptions, "options"  , 2, N_("read options from file")},
  { oDebug,	"debug"     ,4|16, N_("set debugging flags")},
  { oDebugAll, "debug-all" ,0, N_("enable full debugging")},
  { oDebugWait,"debug-wait",1, "@"},
  { oNoDetach, "no-detach" ,0, N_("do not detach from the console")},
  { oNoGrab, "no-grab"     ,0, N_("do not grab keyboard and mouse")},
  { oClient, "client"      ,0, N_("run in client mode for testing")},
  { oLogFile, "log-file"   ,2, N_("use a log file for the server")},
  { oShutdown, "shutdown"  ,0, N_("shutdown the agent")},
  { oFlush   , "flush"     ,0, N_("flush the cache")},
  { oBatch   , "batch"     ,0, N_("run without asking a user")},
  { oLDAPFile, "ldapserverlist-file", 2, N_("File containting list of LDAP servers for certificate and CRL lookup")},
  { oLDAPAddServers, "add-servers", 0, N_("Add new servers discovered in CRL distribution points to serverlist")},
  { oLDAPTimeout, "ldaptimeout", 1, N_("Number of seconds before LDAP requests time out")},
  { oMaxReplies, "max-replies", 1, N_("Maximum number of certificates to return in one query")},
  { oListCRLs, "list-crls", 0, N_("List the contents of the CRL cache")},
  { oLoadCRL, "load-crl", 2, N_("Load CRL from file into cache")},
  {0}
};



static const char *
my_strusage( int level )
{
    const char *p;
    switch( level ) {
      case 11: p = "dirmngr";
	break;
      case 13: p = VERSION; break;
      case 14: p = "Copyright (C) 2002 Klarlvdalens Datakonsult AB"; break;
      case 17: p = PRINTABLE_OS_NAME; break;
      case 19: p =
	    _("Please report bugs to <gpa-dev@gnupg.org>.\n");
	break;
      case 1:
      case 40:	p =
	    _("Usage: dirmngr [options] (-h for help)");
	break;
      case 41:	p =
	    _("Syntax: dirmngr [options] [command [args]]\n"
	      "LDAP access for GnuPG\n");
	break;

      default:	p = NULL;
    }
    return p;
}

static struct ldap_server_s*
parse_ldapserver_file( const char* filename )
{
  static char buffer[1024];
  struct ldap_server_s* serverstart = 0;  
  struct ldap_server_s* servers = 0;  
  struct ldap_server_s server;
  char* bufp = buffer;
  int c;
  int count = 0;
  FILE* fp;

  fp = fopen( filename, "r" );
  if( !fp ) return NULL;

  memset( &server, 0, sizeof( server ) );
  while( (c = getc( fp ) ) != EOF ) {
    struct ldap_server_s* tmp_server;
    if( bufp == buffer ) {
      if( c == '\n' ) continue; /* eat empty lines */
      else if( c == '#' ) {
	/* comment */
	while( (c = getc( fp ) ) != EOF ) {
	  if( c == '\n' ) {
	    break;
	  }
	}
	if( c == '\n' ) continue;
      }
    }
    if( c == '\n' ) {
      char* tmp;
      *bufp = 0;
      bufp = buffer;
      /* host:port:user:pass:base */
      /* host */
      tmp = strchr( bufp, ':' );
      if( tmp ) {
	*tmp = 0;
	server.host = xstrdup( bufp );
	*tmp = ':';
	bufp = tmp+1;
      } else goto get_out;
      /* port */
      tmp = strchr( bufp, ':' );
      if( tmp ) {
	*tmp = 0;
	server.port = atoi( bufp );
	*tmp = ':';
	bufp = tmp+1;
      } else goto get_out;
      /* user */
      tmp = strchr( bufp, ':' );
      if( tmp ) {
	*tmp = 0;
	server.user = xstrdup( bufp );
	*tmp = ':';
	bufp = tmp+1;
      } else goto get_out;
      /* pass */
      tmp = strchr( bufp, ':' );
      if( tmp ) {
	*tmp = 0;
	server.pass = xstrdup( bufp );
	*tmp = ':';
	bufp = tmp+1;
      } else goto get_out;
      /* base */
      tmp = strchr( bufp, '\0' );
      if( tmp ) {
	*tmp = 0;
	server.base = xstrdup( bufp );
      } else goto get_out;
    get_out:
      log_debug( "got host=%s, port=%d, user=%s, pass=%s, base=%s\n", 
		 server.host, server.port, server.user, server.pass, 
		 server.base );
      ++count;
      bufp = buffer;
      tmp_server = xmalloc( sizeof( struct ldap_server_s ) );
      memcpy( tmp_server, &server, sizeof( server ) );
      memset( &server, 0, sizeof( server ) );
      if( serverstart == NULL ) {
	serverstart = tmp_server;
	servers = tmp_server;
      } else {
	servers->next = tmp_server;
	servers = tmp_server;
      }      
    } else {
      *(bufp++) = c;
    }
  }
  return serverstart;
}

static void free_ldapservers_list( struct ldap_server_s* servers )
{
  if( servers == NULL ) return;
  if( servers->next != NULL ) free_ldapservers_list( servers->next );
  if( servers != NULL ) xfree( servers );
}

/* Used by gcry for logging */
static void
my_gcry_logger (void *dummy, int level, const char *fmt, va_list arg_ptr)
{
  /* translate the log levels */
  switch (level)
    {
    case GCRY_LOG_CONT: level = JNLIB_LOG_CONT; break;
    case GCRY_LOG_INFO: level = JNLIB_LOG_INFO; break;
    case GCRY_LOG_WARN: level = JNLIB_LOG_WARN; break;
    case GCRY_LOG_ERROR:level = JNLIB_LOG_ERROR; break;
    case GCRY_LOG_FATAL:level = JNLIB_LOG_FATAL; break;
    case GCRY_LOG_BUG:  level = JNLIB_LOG_BUG; break;
    case GCRY_LOG_DEBUG:level = JNLIB_LOG_DEBUG; break;
    default:            level = JNLIB_LOG_ERROR; break;  
    }
  log_logv (level, fmt, arg_ptr);
}

void dirmgr_exit( int rc )
{
  exit( rc );
}

int
main (int argc, char **argv )
{
  ARGPARSE_ARGS pargs;
  int orig_argc;
  char **orig_argv;
  FILE *configfp = NULL;
  char *configname = NULL;
  const char *shell;
  unsigned configlineno;
  int parse_debug = 0;
  int default_config =1;
  int greeting = 0;
  int nogreeting = 0;
  int server_mode = 1; /* fixme: this is the default for now */
  int client = 0;
  int do_shutdown = 0;
  int do_flush = 0;
  int nodetach = 0;
  int csh_style = 0;
  char *logfile = NULL;
  char *ldapfile = NULL;
  int debug_wait = 0;
  int list_crls = 0;
  char* load_crl_filename = NULL;

  set_strusage (my_strusage);
  log_set_prefix ("dirmngr", 1|4); 

  /* check that the libraries are suitable.  Do it here because
     the option parsing may need services of the library */
  if (!gcry_check_version ( "1.1.7" ) )
    {
      log_fatal( _("libgcrypt is too old (need %s, have %s)\n"),
                 VERSION, gcry_check_version (NULL) );
    }

  ksba_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free );
  assuan_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free);
  gcry_set_log_handler (my_gcry_logger, NULL);

  shell = getenv ("SHELL");
  if (shell && strlen (shell) >= 3 && !strcmp (shell+strlen (shell)-3, "csh") )
    csh_style = 1;
  
  dirmngr_opt.homedir = getenv("GNUPGHOME");
  if (!dirmngr_opt.homedir || !*dirmngr_opt.homedir)
    {
#ifdef HAVE_DRIVE_LETTERS
      dirmngr_opt.homedir = "c:/gnupg";
#else
      dirmngr_opt.homedir = "~/.gnupg";
#endif
    }

  /* LDAP defaults */
  dirmngr_opt.add_new_ldapservers = 0; /* disabled by default */
  dirmngr_opt.ldaptimeout.tv_sec = 100; /* arbitrary large timeout */
  dirmngr_opt.ldaptimeout.tv_usec = 0;

  /* Default max replies */
  dirmngr_opt.max_replies = 10;

  /* check whether we have a config file on the commandline */
  orig_argc = argc;
  orig_argv = argv;
  pargs.argc = &argc;
  pargs.argv = &argv;
  pargs.flags= 1|(1<<6);  /* do not remove the args, ignore version */
  while (arg_parse( &pargs, opts))
    {
      if (pargs.r_opt == oDebug || pargs.r_opt == oDebugAll)
        parse_debug++;
      else if (pargs.r_opt == oOptions)
        { /* yes there is one, so we do not try the default one, but
	     read the option file when it is encountered at the
	     commandline */
          default_config = 0;
	}
	else if (pargs.r_opt == oNoOptions)
          default_config = 0; /* --no-options */
	else if (pargs.r_opt == oHomedir)
          dirmngr_opt.homedir = pargs.r.ret_str;
    }

  if (default_config)
    configname = make_filename (dirmngr_opt.homedir, "dirmngr.conf", NULL );
  
  argc = orig_argc;
  argv = orig_argv;
  pargs.argc = &argc;
  pargs.argv = &argv;
  pargs.flags=  1;  /* do not remove the args */
 next_pass:
  if (configname)
    {
      configlineno = 0;
      configfp = fopen (configname, "r");
      if (!configfp)
        {
          if (default_config)
            {
              if( parse_debug )
                log_info (_("NOTE: no default option file `%s'\n"),
                          configname );
	    }
          else
            {
              log_error (_("option file `%s': %s\n"),
                         configname, strerror(errno) );
              exit(2);
	    }
          xfree (configname); 
          configname = NULL;
	}
      if (parse_debug && configname )
        log_info (_("reading options from `%s'\n"), configname );
      default_config = 0;
    }

  while (optfile_parse( configfp, configname, &configlineno, &pargs, opts) )
    {
      switch (pargs.r_opt)
        {
        case oQuiet: dirmngr_opt.quiet = 1; break;
        case oVerbose: dirmngr_opt.verbose++; break;
        case oBatch: dirmngr_opt.batch=1; break;

        case oDebug: dirmngr_opt.debug |= pargs.r.ret_ulong; break;
        case oDebugAll: dirmngr_opt.debug = ~0; break;
        case oDebugWait: debug_wait = pargs.r.ret_int; break;

        case oOptions:
          /* config files may not be nested (silently ignore them) */
          if (!configfp)
            {
		xfree(configname);
		configname = xstrdup(pargs.r.ret_str);
		goto next_pass;
	    }
          break;
        case oNoGreeting: nogreeting = 1; break;
        case oNoVerbose: dirmngr_opt.verbose = 0; break;
        case oNoOptions: break; /* no-options */
        case oHomedir: dirmngr_opt.homedir = pargs.r.ret_str; break;
        case oNoDetach: nodetach = 1; break;
        case oClient: client = 1; break;
        case oShutdown: client = 1; do_shutdown = 1; break;
        case oFlush: client = 1; do_flush = 1; break;
        case oLogFile: logfile = pargs.r.ret_str; break;
        case oCsh: csh_style = 1; break;
        case oSh: csh_style = 0; break;
        case oServer: server_mode = 1; break;
	case oLDAPFile: ldapfile = pargs.r.ret_str; break;
	case oLDAPAddServers: dirmngr_opt.add_new_ldapservers = 1; break;
	case oLDAPTimeout: 
	  dirmngr_opt.ldaptimeout.tv_sec = pargs.r.ret_int; 
	  dirmngr_opt.ldaptimeout.tv_usec = 0;
	  break;
	case oMaxReplies: dirmngr_opt.max_replies = pargs.r.ret_int; break;
	case oListCRLs: list_crls = 1; break;
	case oLoadCRL: load_crl_filename = pargs.r.ret_str; break;
        default : pargs.err = configfp? 1:2; break;
	}
    }
  if (configfp)
    {
      fclose( configfp );
      configfp = NULL;
      xfree(configname);
      configname = NULL;
      goto next_pass;
    }
  xfree (configname);
  configname = NULL;
  if (log_get_errorcount(0))
    exit(2);
  if (nogreeting )
    greeting = 0;

  if (greeting)
    {
      fprintf (stderr, "%s %s; %s\n",
                 strusage(11), strusage(13), strusage(14) );
      fprintf (stderr, "%s\n", strusage(15) );
    }

  /* LDAP server list form file */
  if( !ldapfile ) {
    ldapfile = make_filename( dirmngr_opt.homedir, "dirmngr_ldapservers.conf", NULL );
  }

  dirmngr_opt.ldapservers = parse_ldapserver_file( ldapfile );

#ifdef IS_DEVELOPMENT_VERSION
  log_info ("NOTE: this is a development version!\n");
#endif

/*    socket_name =  make_filename (dirmngr_opt.homedir, "S.dirmngr", NULL ); */
/*    if (strchr ( socket_name, ':') ) */
/*      { */
/*        log_error ("colons are not allowed in the socket name\n"); */
/*        exit (1); */
/*      } */
   
  if( list_crls ) {
    /* Just list the CRL cache and exit */
    CrlCache* c = crl_cache_init();
    crl_cache_list( c, stdout );
    free_ldapservers_list( dirmngr_opt.ldapservers );
    return 0;
  }

  if( load_crl_filename ) {
    CrlCache* c = crl_cache_init();
    int rc = crl_cache_load( c, load_crl_filename, NULL );
    free_ldapservers_list( dirmngr_opt.ldapservers );
    return rc;
  }

  if (server_mode)
    { /* for now this is the simple pipe based server */
      if (logfile)
        {
          log_set_file (logfile);
          log_set_prefix (NULL, 2|4);
        }
       
/*        if ( atexit (cleanup) ) */
/*          { */
/*            log_error ("atexit failed\n"); */
/*            cleanup (); */
/*            exit (1); */
/*          } */

      if (debug_wait)
        {
          log_debug ("waiting for debugger - my pid is %u .....\n",
                     (unsigned int)getpid());
          sleep (debug_wait);
          log_debug ("... okay\n");
         }
      /* FIXME: start_command_handler (); */
      start_command_handler();
    }
  
  free_ldapservers_list( dirmngr_opt.ldapservers );
  return 0;
}


